/* 
 * x86_64.c
 *
 * Copyright (C) 2006, 2007  NEC Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */
#ifdef __x86_64__

#include "makedumpfile.h"

/*
 *  Include both vmalloc'd and module address space as VMALLOC space.
 */
int
is_vmalloc_addr(ulong vaddr)
{
	return ((vaddr >= VMALLOC_START && vaddr <= VMALLOC_END) ||
	    (vaddr >= MODULES_VADDR && vaddr <= MODULES_END));
}

int
get_phys_base_x86_64()
{
	int i;
	struct pt_load_segment *pls;

	/*
	 * Get the relocatable offset
	 */
	info->phys_base = 0; /* default/traditional */

	for (i = 0; i < info->num_load_memory; i++) {
		pls = &info->pt_load_segments[i];
		if ((pls->virt_start >= __START_KERNEL_map) &&
		    !(is_vmalloc_addr(pls->virt_start))) {

			info->phys_base = pls->phys_start -
			    (pls->virt_start & ~(__START_KERNEL_map));

			break;
		}
	}

	return TRUE;
}

int
get_machdep_info_x86_64()
{
	info->section_size_bits = _SECTION_SIZE_BITS;
	info->max_physmem_bits  = _MAX_PHYSMEM_BITS;

	return TRUE;
}

off_t
vaddr_to_offset_x86_64(unsigned long vaddr)
{
	int i;
	off_t offset;
	unsigned long paddr, phys_base;
	struct pt_load_segment *pls;

	/*
	 * Note this computation of phys_base works for RHEL5
	 * because we always keep phys_base at 0x200000.  This 
	 * doesn't work in general 
	 */
	phys_base = info->phys_base;

	if (vaddr >= __START_KERNEL_map)
		paddr = vaddr - __START_KERNEL_map + phys_base;
	else
		paddr = vaddr - PAGE_OFFSET;

	for (i = offset = 0; i < info->num_load_memory; i++) {
		pls = &info->pt_load_segments[i];
		if ((paddr >= pls->phys_start)
		    && (paddr < pls->phys_end)) {
			offset = (off_t)(paddr - pls->phys_start) +
				pls->file_offset;
				break;
		}
	}
	return offset;
}

/*
 * for Xen extraction
 */
unsigned long long
kvtop_xen_x86_64(unsigned long kvaddr)
{
	unsigned long long dirp, entry;

	if (!is_xen_vaddr(kvaddr)){
		ERRMSG("!is_xen_vaddr(kvaddr=%lx)\n",kvaddr);
		return 0;
	}

	if (is_xen_text(kvaddr)){
		return (unsigned long)kvaddr - XEN_VIRT_START + info->xen_phys_start;
	}

	if (is_direct(kvaddr)){
		return (unsigned long)kvaddr - DIRECTMAP_VIRT_START;
	}

	dirp = kvtop_xen_x86_64(SYMBOL(pgd_l4));
	dirp += pml4_index(kvaddr) * sizeof(unsigned long long);
	if (!readmem(PADDR, dirp, &entry, sizeof(entry)))
		return 0;

	if (!(entry & _PAGE_PRESENT))
		return 0;
	dirp = entry & ENTRY_MASK;
	dirp += pgd_index(kvaddr) * sizeof(unsigned long long);
	if (!readmem(PADDR, dirp, &entry, sizeof(entry)))
		return 0;
 
	if (!(entry & _PAGE_PRESENT))
		return 0;
	dirp = entry & ENTRY_MASK;
	dirp += pmd_index(kvaddr) * sizeof(unsigned long long);
	if (!readmem(PADDR, dirp, &entry, sizeof(entry)))
		return 0;

	if (!(entry & _PAGE_PRESENT))
		return 0;
	if (entry & _PAGE_PSE) {
		entry = (entry & ENTRY_MASK) + (kvaddr & ((1UL << PMD_SHIFT) - 1));
		return entry;
	}
	dirp = entry & ENTRY_MASK;
	dirp += pte_index(kvaddr) * sizeof(unsigned long long);
	if (!readmem(PADDR, dirp, &entry, sizeof(entry)))
		return 0;

	if (!(entry & _PAGE_PRESENT)) {
		return 0;
	}

	entry = (entry & ENTRY_MASK) + (kvaddr & ((1UL << PTE_SHIFT) - 1));

	return entry;
}

int get_xen_info_x86_64()
{
	unsigned long frame_table_vaddr;
	unsigned long xen_end;
	int i;

	if (SYMBOL(pgd_l4) == NOT_FOUND_SYMBOL) {
		ERRMSG("Can't get pml4.\n");
		return FALSE;
	}

	if (SYMBOL(frame_table) == NOT_FOUND_SYMBOL) {
		ERRMSG("Can't get the symbol of frame_table.\n");
		return FALSE;
	}
	if (!readmem(VADDR_XEN, SYMBOL(frame_table), &frame_table_vaddr,
	    sizeof(frame_table_vaddr))) {
		ERRMSG("Can't get the value of frame_table.\n");
		return FALSE;
	}
	info->frame_table_vaddr = frame_table_vaddr;

	if (SYMBOL(xenheap_phys_end) == NOT_FOUND_SYMBOL) {
		ERRMSG("Can't get the symbol of xenheap_phys_end.\n");
		return FALSE;
	}
	if (!readmem(VADDR_XEN, SYMBOL(xenheap_phys_end), &xen_end,
	    sizeof(xen_end))) {
		ERRMSG("Can't get the value of xenheap_phys_end.\n");
		return FALSE;
	}
	info->xen_heap_end = (xen_end >> PAGESHIFT());
	info->xen_heap_start = 0;

	/*
	 * pickled_id == domain addr for x86_64
	 */
	for (i = 0; i < info->num_domain; i++) {
		info->domain_list[i].pickled_id =
			info->domain_list[i].domain_addr;
	}

	return TRUE;
}

#endif /* x86_64 */

